@@ -3,6 +3,7 @@
  * alsa.c -- The Advanced Linux Sound System backend for Speech Dispatcher
  *
  * Copyright (C) 2005,2006 Brailcom, o.p.s.
+ * Copyright (C) 2019-2024 Samuel Thibault <samuel.thibault@ens-lyon.org>
  *
  * This is free software; you can redistribute it and/or modify it under the
  * terms of the GNU Lesser General Public License as published by the Free
@@ -15,9 +16,7 @@
  * General Public License for more details.
  *
  * You should have received a copy of the GNU Lesser General Public License
- * along with this package; see the file COPYING.  If not, write to the Free
- * Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
- * 02110-1301, USA.
+ * along with this program.  If not, see <https://www.gnu.org/licenses/>.
  *
  * $Id: alsa.c,v 1.30 2008-10-15 17:27:32 hanke Exp $
  */
@@ -36,10 +35,17 @@
 #include <glib.h>
 
 #include <alsa/asoundlib.h>
+#include <alsa/pcm.h>
 
+#ifdef USE_DLOPEN
+#define SPD_AUDIO_PLUGIN_ENTRY spd_audio_plugin_get
+#else
 #define SPD_AUDIO_PLUGIN_ENTRY spd_alsa_LTX_spd_audio_plugin_get
+#endif
 #include <spd_audio_plugin.h>
 
+#include "../common/common.h"
+
 typedef struct {
 	AudioID id;
 	snd_pcm_t *alsa_pcm;	/* identifier of the ALSA device */
@@ -48,7 +54,9 @@ typedef struct {
 	snd_pcm_uframes_t alsa_buffer_size;
 	pthread_mutex_t alsa_pcm_mutex;	/* mutex to guard the state of the device */
 	pthread_mutex_t alsa_pipe_mutex;	/* mutex to guard the stop pipes */
+	pthread_cond_t alsa_pipe_cond;	/* mutex to guard the stop pipes */
 	int alsa_stop_pipe[2];	/* Pipe for communication about stop requests */
+	int stop_requested;	/* Whether we want to stop */
 	int alsa_fd_count;	/* Counter of descriptors to poll */
 	struct pollfd *alsa_poll_fds;	/* Descriptors to poll */
 	int alsa_opened;	/* 1 between snd_pcm_open and _close, 0 otherwise */
@@ -77,39 +85,8 @@ static int wait_for_poll(spd_alsa_id_t * id, struct pollfd *alsa_poll_fds,
 #endif
 
 /* Put a message into the logfile (stderr) */
-#define MSG(level, arg...) \
-	if(level <= alsa_log_level){ \
-		time_t t; \
-		struct timeval tv; \
-		char *tstr; \
-		t = time(NULL); \
-		tstr = g_strdup(ctime(&t)); \
-		tstr[strlen(tstr)-1] = 0; \
-		gettimeofday(&tv,NULL); \
-		fprintf(stderr," %s [%d]",tstr, (int) tv.tv_usec); \
-		fprintf(stderr," ALSA: "); \
-		fprintf(stderr,arg); \
-		fprintf(stderr,"\n"); \
-		fflush(stderr); \
-		g_free(tstr); \
-	}
-
-#define ERR(arg...) \
-	{ \
-		time_t t; \
-		struct timeval tv; \
-		char *tstr; \
-		t = time(NULL); \
-		tstr = g_strdup(ctime(&t)); \
-		tstr[strlen(tstr)-1] = 0; \
-		gettimeofday(&tv,NULL); \
-		fprintf(stderr," %s [%d]",tstr, (int) tv.tv_usec); \
-		fprintf(stderr," ALSA ERROR: "); \
-		fprintf(stderr,arg); \
-		fprintf(stderr,"\n"); \
-		fflush(stderr); \
-		g_free(tstr); \
-	}
+#define MSG(level, arg, ...) if (level <= alsa_log_level) { MSG(0, "ALSA: " arg, ##__VA_ARGS__); }
+#define ERR(arg, ...) MSG(0, "ALSA ERROR: " arg, ##__VA_ARGS__)
 
 static int alsa_log_level;
 static char const *alsa_play_cmd = "aplay";
@@ -218,14 +195,18 @@ static int _alsa_close(spd_alsa_id_t * id)
 
 	MSG(1, "Closing ALSA device");
 
-	if (id->alsa_opened == 0)
-		return 0;
-
 	pthread_mutex_lock(&id->alsa_pipe_mutex);
+
+	if (id->alsa_opened == 0) {
+		pthread_mutex_unlock(&id->alsa_pipe_mutex);
+		return 0;
+	}
+
 	id->alsa_opened = 0;
 
 	if ((err = snd_pcm_close(id->alsa_pcm)) < 0) {
 		MSG(2, "Cannot close ALSA device (%s)", snd_strerror(err));
+		pthread_mutex_unlock(&id->alsa_pipe_mutex);
 		return -1;
 	}
 
@@ -260,6 +241,7 @@ static AudioID *alsa_open(void **pars)
 	alsa_id = (spd_alsa_id_t *) g_malloc(sizeof(spd_alsa_id_t));
 
 	pthread_mutex_init(&alsa_id->alsa_pipe_mutex, NULL);
+	pthread_cond_init(&alsa_id->alsa_pipe_cond, NULL);
 
 	alsa_id->alsa_opened = 0;
 
@@ -366,51 +348,29 @@ int wait_for_poll(spd_alsa_id_t * id, struct pollfd *alsa_poll_fds,
 
 		/* Is ALSA ready for more input? */
 		if ((revents & POLLOUT)) {
-			// MSG("Poll: Ready for more input");
+			MSG(5, "Poll: Ready for more input");
 			return 0;
 		}
 	}
 }
 
-#define ERROR_EXIT()\
+#define ERROR_EXIT() do {\
 	g_free(track_volume.samples); \
 	ERR("alsa_play() abnormal exit"); \
 	_alsa_close(alsa_id); \
-	return -1;
+	return -1; \
+} while (0)
 
-/* Play the track _track_ (see spd_audio.h) using the id->alsa_pcm device and
- id-hw_params parameters. This is a blocking function, however, it's possible
- to interrupt playing from a different thread with alsa_stop(). alsa_play
- returns after and immediatelly after the whole sound was played on the
- speakers.
-
- The idea is that we get the ALSA file descriptors and we will poll() to see
- when alsa is ready for more input while sleeping in the meantime. We will
- additionally poll() for one more descriptor used by alsa_stop() to notify the
- thread with alsa_play() that the stop of the playback is requested. The
- variable can_be_stopped is used for very simple synchronization between the
- two threads. */
-static int alsa_play(AudioID * id, AudioTrack track)
+/* Configure ALSA playback for the given configuration of track
+   But do not play anything yet */
+static int alsa_begin(AudioID * id, AudioTrack track)
 {
 	snd_pcm_format_t format;
-	int bytes_per_sample;
-	int num_bytes;
 	spd_alsa_id_t *alsa_id = (spd_alsa_id_t *) id;
 
-	signed short *output_samples;
-
-	AudioTrack track_volume;
-	float real_volume;
-	int i;
-
 	int err;
-	int ret;
 
-	snd_pcm_uframes_t framecount;
 	snd_pcm_uframes_t period_size;
-	size_t samples_per_period;
-	size_t silent_samples;
-	size_t volume_size;
 	unsigned int sr;
 
 	snd_pcm_state_t state;
@@ -457,6 +417,7 @@ static int alsa_play(AudioID * id, AudioTrack track)
 		pthread_mutex_unlock(&alsa_id->alsa_pipe_mutex);
 		return -1;
 	}
+	alsa_id->stop_requested = 0;
 
 	/* Find how many descriptors we will get for poll() */
 	alsa_id->alsa_fd_count =
@@ -510,9 +471,7 @@ static int alsa_play(AudioID * id, AudioTrack track)
 			ERR("unknown audio format (%d)", alsa_id->id.format);
 			return -1;
 		}
-		bytes_per_sample = 2;
 	} else if (track.bits == 8) {
-		bytes_per_sample = 1;
 		format = SND_PCM_FORMAT_S8;
 	} else {
 		ERR("Unsupported sound data format, track.bits = %d",
@@ -585,7 +544,7 @@ static int alsa_play(AudioID * id, AudioTrack track)
 		    snd_strerror(err));
 		return -1;
 	}
-	MSG(4, "Buffer size on ALSA device is %d bytes",
+	MSG(4, "Buffer size on ALSA device is %d frames",
 	    (int)alsa_id->alsa_buffer_size);
 
 	/* This is probably better left for the device driver to decide */
@@ -598,15 +557,8 @@ static int alsa_play(AudioID * id, AudioTrack track)
 
 	/* Get period size. */
 	snd_pcm_hw_params_get_period_size(alsa_id->alsa_hw_params, &period_size,
-					  0);
-
-	/* Calculate size of silence at end of buffer. */
-	samples_per_period = period_size * track.num_channels;
-	//    MSG("samples per period = %i", samples_per_period);
-	//    MSG("num_samples = %i", track.num_samples);
-	silent_samples =
-	    samples_per_period - (track.num_samples % samples_per_period);
-	//    MSG("silent samples = %i", silent_samples);
+	                                  0);
+	MSG(4, "Period size on ALSA device is %lu frames", (unsigned long) period_size);
 
 	MSG(4, "Preparing device for playback");
 	if ((err = snd_pcm_prepare(alsa_id->alsa_pcm)) < 0) {
@@ -616,8 +568,34 @@ static int alsa_play(AudioID * id, AudioTrack track)
 		return -1;
 	}
 
+	return 0;
+}
+
+/* Push audio track to ALSA playback */
+static int alsa_feed(AudioID * id, AudioTrack track)
+{
+	int bytes_per_sample;
+	int num_bytes;
+	spd_alsa_id_t *alsa_id = (spd_alsa_id_t *) id;
+
+	AudioTrack track_volume;
+	float real_volume;
+	int i;
+
+	signed short *output_samples;
+
+	int err;
+	int ret;
+
+	snd_pcm_state_t state;
+
+	snd_pcm_uframes_t framecount;
+	size_t volume_size;
+
+	bytes_per_sample = track.bits / 8;
+
 	/* Calculate space needed to round up to nearest period size. */
-	volume_size = bytes_per_sample * (track.num_samples + silent_samples);
+	volume_size = bytes_per_sample * track.num_samples;
 	MSG(4, "volume size = %i", (int)volume_size);
 
 	/* Create a copy of track with adjusted volume. */
@@ -628,44 +606,14 @@ static int alsa_play(AudioID * id, AudioTrack track)
 	for (i = 0; i <= track.num_samples - 1; i++)
 		track_volume.samples[i] = track.samples[i] * real_volume;
 
-	if (silent_samples > 0) {
-		u_int16_t silent16;
-		u_int8_t silent8;
-
-		/* Fill remaining space with silence */
-		MSG(4,
-		    "Filling with silence up to the period size, silent_samples=%d",
-		    (int)silent_samples);
-		/* TODO: This hangs.  Why?
-		   snd_pcm_format_set_silence(format,
-		   track_volume.samples + (track.num_samples * bytes_per_sample), silent_samples);
-		 */
-		switch (bytes_per_sample) {
-		case 2:
-			silent16 = snd_pcm_format_silence_16(format);
-			for (i = 0; i < silent_samples; i++)
-				track_volume.samples[track.num_samples + i] =
-				    silent16;
-			break;
-		case 1:
-			silent8 = snd_pcm_format_silence(format);
-			for (i = 0; i < silent_samples; i++)
-				track_volume.samples[track.num_samples + i] =
-				    silent8;
-			break;
-		}
-	}
-
 	/* Loop until all samples are played on the device. */
 	output_samples = track_volume.samples;
-	num_bytes = (track.num_samples + silent_samples) * bytes_per_sample;
-	//    MSG("Still %d bytes left to be played", num_bytes);
+	num_bytes = volume_size;
+	MSG(4, "%d bytes to be played", num_bytes);
 	while (num_bytes > 0) {
 
 		/* Write as much samples as possible */
 		framecount = num_bytes / bytes_per_sample / track.num_channels;
-		if (framecount < period_size)
-			framecount = period_size;
 
 		/* Report current state state */
 		state = snd_pcm_state(alsa_id->alsa_pcm);
@@ -676,17 +624,22 @@ static int alsa_play(AudioID * id, AudioTrack track)
 		ret =
 		    snd_pcm_writei(alsa_id->alsa_pcm, output_samples,
 				   framecount);
-		//        MSG("Sent %d of %d remaining bytes", ret*bytes_per_sample, num_bytes);
+		if (ret >= 0)
+			MSG(5, "Sent %d of %d remaining bytes", ret*bytes_per_sample*track.num_channels, num_bytes);
 
 		if (ret == -EAGAIN) {
 			MSG(4, "Warning: Forced wait!");
 			snd_pcm_wait(alsa_id->alsa_pcm, 100);
 		} else if (ret == -EPIPE) {
+			MSG(4, "Warning: returned EPIPE!");
 			if (xrun(alsa_id) != 0)
 				ERROR_EXIT();
+#ifdef ESTRPIPE
 		} else if (ret == -ESTRPIPE) {
+			MSG(4, "Warning: returned ESTRPIPE!");
 			if (suspend(alsa_id) != 0)
 				ERROR_EXIT();
+#endif
 		} else if (ret == -EBUSY) {
 			MSG(4, "WARNING: sleeping while PCM BUSY");
 			usleep(100);
@@ -728,6 +681,7 @@ static int alsa_play(AudioID * id, AudioTrack track)
 				return -1;
 			}
 
+			/* Terminating (successfully or after a stop) */
 			goto terminate;
 		}
 
@@ -738,14 +692,28 @@ static int alsa_play(AudioID * id, AudioTrack track)
 		/* Stop requests can be issued again */
 	}
 
-	MSG(4, "Draining...");
+terminate:
+	if (track_volume.samples != NULL)
+		g_free(track_volume.samples);
 
-	/* We want to next "device ready" notification only after the buffer is
-	   already empty */
+	return 0;
+}
+
+/* Drain ALSA playback until only `left' samples are left in the buffer */
+static int alsa_drain_left(AudioID * id, snd_pcm_uframes_t left)
+{
+	spd_alsa_id_t *alsa_id = (spd_alsa_id_t *) id;
+
+	int err;
+
+	MSG(4, "Draining until %lu frames left...", (unsigned long) left);
+
+	/* We want to get next "device ready" notification only after the buffer
+	   is already empty */
 	err =
 	    snd_pcm_sw_params_set_avail_min(alsa_id->alsa_pcm,
 					    alsa_id->alsa_sw_params,
-					    alsa_id->alsa_buffer_size);
+					    alsa_id->alsa_buffer_size - left);
 	if (err < 0) {
 		ERR("Unable to set avail min for playback: %s\n",
 		    snd_strerror(err));
@@ -776,12 +744,137 @@ static int alsa_play(AudioID * id, AudioTrack track)
 			return -1;
 		}
 	}
+
+	/* When ALSA is going through Pulseaudio, wait_for_poll returns too
+	   early because the file descriptor is always available for writing
+	   :/ */
+	while (!alsa_id->stop_requested)
+	{
+		snd_pcm_sframes_t frames;
+		snd_pcm_state_t state;
+		struct timeval tv;
+		struct timespec ts;
+
+		/* Poll server */
+		frames = snd_pcm_avail(alsa_id->alsa_pcm);
+		if (frames < 0) {
+			MSG(4, "Drain: Buffer clear");
+			break;
+		}
+
+		MSG(5, "Drain: %lu frames left in buffer",
+			(unsigned long) alsa_id->alsa_buffer_size - frames);
+		if (alsa_id->alsa_buffer_size - frames <= left) {
+			MSG(4, "Drain: Buffer clear enough");
+			break;
+		}
+
+		state = snd_pcm_state(alsa_id->alsa_pcm);
+		if (err != 0) {
+			MSG(4, "Drain: Status error %d", err);
+			break;
+		}
+		if (state == SND_PCM_STATE_XRUN) {
+			MSG(4, "Drain: Playback terminated");
+			break;
+		}
+
+		/* Poll every 10ms */
+		gettimeofday(&tv, NULL);
+		ts.tv_sec = tv.tv_sec;
+		ts.tv_nsec = tv.tv_usec * 1000 + 10000000;
+		if (ts.tv_nsec >= 1000000000) {
+			ts.tv_sec += 1;
+			ts.tv_nsec -= 1000000000;
+		}
+
+		pthread_mutex_lock(&alsa_id->alsa_pipe_mutex);
+		if (!alsa_id->stop_requested)
+			pthread_cond_timedwait(&alsa_id->alsa_pipe_cond,
+					       &alsa_id->alsa_pipe_mutex, &ts);
+		pthread_mutex_unlock(&alsa_id->alsa_pipe_mutex);
+	}
+
 	MSG(4, "Draining terminated");
 
-terminate:
-	/* Terminating (successfully or after a stop) */
-	if (track_volume.samples != NULL)
-		g_free(track_volume.samples);
+	return 0;
+}
+
+/* Drain until the amount of samples left in the buffer is big enough to make
+   sure we will have time to push the rest of the audio */
+static int alsa_drain_overlap(AudioID * id, AudioTrack track)
+{
+	/* We typically want 20ms overlap: usually about a period size, and
+	   small enough to be unnoticeable */
+	unsigned min_ms = 20;
+
+	spd_alsa_id_t *alsa_id = (spd_alsa_id_t *) id;
+	snd_pcm_uframes_t min;
+	snd_pcm_uframes_t period_size;
+	snd_pcm_uframes_t min2;
+	snd_pcm_uframes_t left;
+
+	min = (min_ms * track.sample_rate) / 1000;
+
+	/* Get period size. */
+	snd_pcm_hw_params_get_period_size(alsa_id->alsa_hw_params, &period_size,
+	                                  0);
+
+	/* Round minimum to two period size */
+	if (2*period_size >= min)
+		min2 = 2*period_size;
+	else
+		min2 = (min + period_size - 1) / period_size;
+
+	/* Usually the buffer size will be far enough */
+	if (alsa_id->alsa_buffer_size >= min2)
+		left = min2;
+	else
+		/* That's very odd. Just wait for some room to be available */
+		left = alsa_id->alsa_buffer_size - 1;
+
+	MSG(4, "Draining with at least %ums left, i.e. %lu frames, with period %lu frames, thus %lu frames, i.e. %lu left of %lu",
+		min_ms, (unsigned long) min, (unsigned long) period_size,
+		(unsigned long) min2, (unsigned long) left, (unsigned long)
+		alsa_id->alsa_buffer_size);
+
+	return alsa_drain_left(id, left);
+}
+
+static int alsa_drain(AudioID * id)
+{
+	return alsa_drain_left(id, 0);
+}
+
+static int alsa_feed_sync(AudioID * id, AudioTrack track)
+{
+	int ret;
+
+	ret = alsa_feed(id, track);
+	if (ret)
+		return ret;
+
+	return alsa_drain(id);
+}
+
+static int alsa_feed_sync_overlap(AudioID * id, AudioTrack track)
+{
+	int ret;
+
+	ret = alsa_feed(id, track);
+	if (ret)
+		return ret;
+
+	return alsa_drain_overlap(id, track);
+}
+
+static int alsa_end(AudioID * id)
+{
+	spd_alsa_id_t *alsa_id = (spd_alsa_id_t *) id;
+	int err;
+
+	if (!alsa_id->stop_requested)
+		alsa_drain(id);
 
 	err = snd_pcm_drop(alsa_id->alsa_pcm);
 	if (err < 0) {
@@ -805,6 +898,33 @@ terminate:
 	return 0;
 }
 
+/* Play the track _track_ (see spd_audio.h) using the id->alsa_pcm device and
+ id-hw_params parameters. This is a blocking function, however, it's possible
+ to interrupt playing from a different thread with alsa_stop(). alsa_play
+ returns after and immediately after the whole sound was played on the
+ speakers.
+
+ The idea is that we get the ALSA file descriptors and we will poll() to see
+ when alsa is ready for more input while sleeping in the meantime. We will
+ additionally poll() for one more descriptor used by alsa_stop() to notify the
+ thread with alsa_play() that the stop of the playback is requested. The
+ variable can_be_stopped is used for very simple synchronization between the
+ two threads. */
+static int alsa_play(AudioID * id, AudioTrack track)
+{
+	int ret;
+
+	ret = alsa_begin(id, track);
+	if (ret)
+		return ret;
+
+	ret = alsa_feed_sync(id, track);
+	if (ret)
+		return ret;
+
+	return alsa_end(id);
+}
+
 #undef ERROR_EXIT
 
 /*
@@ -823,6 +943,8 @@ static int alsa_stop(AudioID * id)
 
 	pthread_mutex_lock(&alsa_id->alsa_pipe_mutex);
 	if (alsa_id->alsa_opened) {
+		alsa_id->stop_requested = 1;
+
 		/* This constant is arbitrary */
 		buf = 42;
 
@@ -831,6 +953,7 @@ static int alsa_stop(AudioID * id)
 			ERR("Can't write stop request to pipe, err %d: %s",
 			    errno, strerror(errno));
 		}
+		pthread_cond_broadcast(&alsa_id->alsa_pipe_cond);
 	}
 	pthread_mutex_unlock(&alsa_id->alsa_pipe_mutex);
 
@@ -869,7 +992,11 @@ static spd_audio_plugin_t alsa_functions = {
 	alsa_close,
 	alsa_set_volume,
 	alsa_set_loglevel,
-	alsa_get_playcmd
+	alsa_get_playcmd,
+	alsa_begin,
+	alsa_feed_sync,
+	alsa_feed_sync_overlap,
+	alsa_end,
 };
 
 spd_audio_plugin_t *alsa_plugin_get(void)
@@ -877,7 +1004,11 @@ spd_audio_plugin_t *alsa_plugin_get(void)
 	return &alsa_functions;
 }
 
-spd_audio_plugin_t *SPD_AUDIO_PLUGIN_ENTRY(void)
-    __attribute__ ((weak, alias("alsa_plugin_get")));
+spd_audio_plugin_t *
+    __attribute__ ((weak))
+    SPD_AUDIO_PLUGIN_ENTRY(void)
+{
+	return &alsa_functions;
+}
 #undef MSG
 #undef ERR
