@@ -3,7 +3,7 @@
  * espeak.c - Speech Dispatcher backend for espeak
  *
  * Copyright (C) 2007 Brailcom, o.p.s.
- * Copyright (C) 2019-2022 Samuel Thibault <samuel.thibault@ens-lyon.org>
+ * Copyright (C) 2019-2024 Samuel Thibault <samuel.thibault@ens-lyon.org>
  *
  * This is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by
@@ -207,6 +207,7 @@ int module_init(char **status_info)
 	    espeak_Initialize(AUDIO_OUTPUT_SYNCHRONOUS, EspeakAudioChunkSize,
 			      NULL, 0);
 #endif
+	DBG(DBG_MODNAME " Sample rate is %d", espeak_sample_rate);
 	if (espeak_sample_rate == EE_INTERNAL_ERROR) {
 		DBG(DBG_MODNAME " Could not initialize engine.");
 		*status_info = g_strdup("Could not initialize engine. ");
@@ -507,7 +508,7 @@ static void espeak_set_punctuation_mode(SPDPunctuation punct_mode)
 	espeak_ERROR ret =
 	    espeak_SetParameter(espeakPUNCTUATION, espeak_punct_mode, 0);
 	if (ret != EE_OK) {
-		DBG(DBG_MODNAME " Failed to set punctuation mode.");
+		DBG(DBG_MODNAME " Failed to set punctuation mode to %d.", espeak_punct_mode);
 	} else {
 		DBG("Set punctuation mode.");
 	}
@@ -531,7 +532,7 @@ static void espeak_set_cap_let_recogn(SPDCapitalLetters cap_mode)
 	espeak_ERROR ret =
 	    espeak_SetParameter(espeakCAPITALS, espeak_cap_mode, 1);
 	if (ret != EE_OK) {
-		DBG(DBG_MODNAME " Failed to set capitals mode.");
+		DBG(DBG_MODNAME " Failed to set capitals mode to %d.", espeak_cap_mode);
 	} else {
 		DBG("Set capitals mode.");
 	}
@@ -543,54 +544,52 @@ static void espeak_set_language_and_voice(char *lang, SPDVoiceType voice_code)
 	DBG(DBG_MODNAME " set_language_and_voice %s %d", lang, voice_code);
 	espeak_ERROR ret;
 
-	unsigned char overlay = 0;
+	espeak_VOICE voice_select;
+	memset(&voice_select, 0, sizeof(voice_select));
+	voice_select.languages = lang;
+
 	switch (voice_code) {
 	case SPD_MALE1:
-		overlay = 0;
+		voice_select.gender = 1;
 		break;
 	case SPD_MALE2:
-		overlay = 1;
+		voice_select.gender = 1;
+		voice_select.variant = 1;
 		break;
 	case SPD_MALE3:
-		overlay = 2;
+		voice_select.gender = 1;
+		voice_select.variant = 2;
 		break;
 	case SPD_FEMALE1:
-		overlay = 11;
+		voice_select.gender = 2;
 		break;
 	case SPD_FEMALE2:
-		overlay = 12;
+		voice_select.gender = 2;
+		voice_select.variant = 1;
 		break;
 	case SPD_FEMALE3:
-		overlay = 13;
+		voice_select.gender = 2;
+		voice_select.variant = 2;
 		break;
 	case SPD_CHILD_MALE:
-		overlay = 4;
+		voice_select.gender = 1;
+		voice_select.age = 10;
 		break;
 	case SPD_CHILD_FEMALE:
-		overlay = 14;
+		voice_select.gender = 2;
+		voice_select.age = 10;
 		break;
 	default:
-		overlay = 0;
 		break;
 	}
 
-	char *name = g_strdup_printf("%s+%d", lang, overlay);
-	DBG(DBG_MODNAME " set_language_and_voice name=%s", name);
-	ret = espeak_SetVoiceByName(name);
+	ret = espeak_SetVoiceByProperties(&voice_select);
 
 	if (ret != EE_OK) {
-		espeak_VOICE voice_select;
-		memset(&voice_select, 0, sizeof(voice_select));
-		voice_select.languages = name;
-		ret = espeak_SetVoiceByProperties(&voice_select);
-	}
-
-	if (ret != EE_OK) {
-		DBG(DBG_MODNAME " Error selecting language %s", name);
+		DBG(DBG_MODNAME " Error selecting language %s", lang);
 	} else {
-		DBG(DBG_MODNAME " Successfully set voice to \"%s\"", name);
+		DBG(DBG_MODNAME " Successfully set voice to \"%s\"", lang);
 	}
-	g_free(name);
 }
 
 static void espeak_set_voice(SPDVoiceType voice)
@@ -678,15 +677,10 @@ static gboolean espeak_send_audio_upto(short *wav, int *sent, int upto)
 	if (wav == NULL || numsamples == 0) {
 		return TRUE;
 	}
-#ifdef ESPEAK_NG_INCLUDE
-	int rate = espeak_ng_GetSampleRate();
-#else
-	int rate = espeak_sample_rate;
-#endif
 	AudioTrack track = {
 		.bits = 16,
 		.num_channels = 1,
-		.sample_rate = rate,
+		.sample_rate = espeak_sample_rate,
 		.num_samples = numsamples,
 		.samples = wav + (*sent),
 	};
@@ -702,6 +696,7 @@ static int synth_callback(short *wav, int numsamples, espeak_EVENT * events)
 	static int numsamples_sent_msg = 0;
 	/* Number of samples already sent during this call to the callback. */
 	int numsamples_sent = 0;
+	int first = 1, nextfirst;
 
 	/* Process server events in case we were told to stop in between */
 	module_process(STDIN_FILENO, 0);
@@ -720,6 +715,7 @@ static int synth_callback(short *wav, int numsamples, espeak_EVENT * events)
 
 	/* Process events and audio data */
 	while (events->type != espeakEVENT_LIST_TERMINATED) {
+		nextfirst = 0;
 		/* Enqueue audio upto event */
 		switch (events->type) {
 		case espeakEVENT_MARK:
@@ -744,6 +740,13 @@ static int synth_callback(short *wav, int numsamples, espeak_EVENT * events)
 			return 1;
 		/* Process actual event */
 		switch (events->type) {
+		case espeakEVENT_SENTENCE:
+		case espeakEVENT_WORD:
+		case espeakEVENT_PHONEME:
+		case espeakEVENT_END:
+			// Ignore
+			break;
+
 		case espeakEVENT_MARK:
 			if (EspeakIndexing) {
 				DBG(DBG_MODNAME " Reporting mark %s", events->id.name);
@@ -764,12 +767,26 @@ static int synth_callback(short *wav, int numsamples, espeak_EVENT * events)
 			// This event never has any audio in the same callback
 			DBG(DBG_MODNAME " Synth terminated");
 			break;
+		case espeakEVENT_SAMPLERATE:
+			DBG(DBG_MODNAME " Got sample rate %d", events->id.number);
+			if (first) {
+				/* espeak-ng currently seems to produce odd
+				 * sample rate changes, so ignore them but the
+				 * initial event for now.
+				 * See https://github.com/espeak-ng/espeak-ng/issues/2028 */
+				espeak_sample_rate = events->id.number;
+				/* Keep looking at sample rate changes until we get something else */
+				nextfirst = 1;
+			}
+			break;
 		default:
+			DBG(DBG_MODNAME " Got unsupported event %d\n", events->type);
 			break;
 		}
 		if (stop_requested)
 			return 1;
 		events++;
+		first = nextfirst;
 	}
 	espeak_send_audio_upto(wav, &numsamples_sent, numsamples);
 	numsamples_sent_msg += numsamples;
@@ -900,7 +917,7 @@ static SPDVoice **espeak_list_synthesis_voices()
 
 			if (!strncmp(identifier, "mb/mb-", 6)) {
 				voicename = g_strdup(identifier + 6);
-				dash = index(voicename, '-');
+				dash = strchr(voicename, '-');
 				if (dash)
 					/* Ignore "-en" language specification */
 					*dash = 0;
